webpackJsonp([19],{428:function(e,t,a){"use strict";function n(e){return e&&e.__esModule?e:{"default":e}}var u=a(3),r=n(u),s=a(4),l=n(s),o=a(5),c=n(o),d=a(7),f=n(d),i=a(6),m=n(i),g=a(1),p=n(g),_=a(9),v=a(11),E=n(v),M=a(572),N=function(e){function t(){return(0,l["default"])(this,t),(0,f["default"])(this,(0,r["default"])(t).apply(this,arguments))}return(0,m["default"])(t,e),(0,c["default"])(t,[{key:"render",value:function(){return p["default"].createElement("div",{className:"container-fluid"},p["default"].createElement("h1",{className:"text-center"},p["default"].createElement(_.FormattedMessage,{id:"errorPage.PageNotFound",defaultMessage:"Page Not Found!"})),p["default"].createElement("img",{className:"img-responsive center-block",src:M}))}}]),t}(E["default"]);e.exports=N},429:function(e,t,a){"use strict";function n(e){return e&&e.__esModule?e:{"default":e}}function u(e){return{state:e}}Object.defineProperty(t,"__esModule",{value:!0});var r=a(17),s=a(428),l=n(s);t["default"]=(0,r.connect)(u)(l["default"])},572:function(e,t,a){e.exports=a.p+"img/_/routes/ErrorPage/assets/images/404.png"}});
//# sourceMappingURL=19.c71778c2489af73b29dd.chunk.js.map